
{*******************************************************}
{                                                       }
{       Borland Delphi Visual Component Library         }
{       TWave and TWavePlayer                           }
{                                                       }
{       Copyright (c) 1998-2000 Vino Rodrigues          }
{       vinorodrigues@yahoo.com                         }
{                                                       }
{       Version History:                                }
{       ----------------                                }
{                                                       }
{       1.1  30-Aug-01                                  }
{            First time I included "version history".   }
{                                                       }
{*******************************************************}

unit Wave;

{$UNDEF DEBUG}
{.$DEFINE DEBUG}

{$IFDEF DEBUG}
  {$D+,C+,O-,Q+,Y+,L+}
{$ELSE}
  {$D-,C-,Q-,Y-,L-}
{$ENDIF}

{$P+,S-,W-,R-,T-,H+,X+}

interface

uses
  SysUtils, Classes, Windows, MMSystem;

const
  RT_WAVE = 'WAVE';

{ TWave }

type
  TWaveOption = (sndLoop, sndNoDefault, sndNoStop, sndNoWait);
  TWaveOptions = set of TWaveOption;

  TWaveSync = (sndSync, sndASync);

  TWave = class(TPersistent)
  private
    FSync: TWaveSync;
    FOptions: TWaveOptions;
    FOnChange: TNotifyEvent;
    FSize: Longint;
    FData: Pointer;
    function Flags: DWORD;
    procedure ReadData(Stream: TStream);
    procedure WriteData(Stream: TStream);
  protected
    property Data: Pointer read FData;
    procedure Changed; dynamic;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure LoadFromFile(const FileName: string);
    procedure LoadFromResourceName(Instance: THandle; const ResName: string; ResType: PChar);
    procedure LoadFromResourceID(Instance: THandle; ResID: Integer; ResType: PChar);
    procedure SaveToFile(const FileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    class function IsSoundInstalled: Boolean;
    function Play: Boolean;
    function PlayFile(const FileName: string): Boolean;
    function PlayResourseName(Instance: THandle; const ResName: string; ResType: PChar): Boolean;
    function PlayResourseID(Instance: THandle; ResID: Integer; ResType: PChar): Boolean;
    function PlayAlias(const AliasName: string): Boolean;
    procedure Stop;
    procedure Clear;
    property Size: Longint read FSize;
    property Sync: TWaveSync read FSync write FSync default sndSync;
    property Options: TWaveOptions read FOptions write FOptions;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

{ TWavePlayer }

  TWavePlayer = class(TComponent)
  private
    FWave: TWave;
    FSync: TWaveSync;
    FOptions: TWaveOptions;
    FBeforePlay: TNotifyEvent;
    FAfterPlay: TNotifyEvent;
    FOnChange: TNotifyEvent;
    procedure SetWave(Value: TWave);
    procedure WaveChange(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    function Play: Boolean;
    function PlayFile(const FileName: string): Boolean;
    function PlayResourseName(Instance: THandle; const ResName: string; ResType: PChar): Boolean;
    function PlayResourseID(Instance: THandle; ResID: Integer; ResType: PChar): Boolean;
    function PlayAlias(const AliasName: string): Boolean;
    procedure Stop;
  published
    property Wave: TWave read FWave write SetWave;
    property Sync: TWaveSync read FSync write FSync default sndSync;
    property Options: TWaveOptions read FOptions write FOptions;
    property BeforePlay: TNotifyEvent read FBeforePlay write FBeforePlay;
    property AfterPlay: TNotifyEvent read FAfterPlay write FAfterPlay;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

{ TWave }

constructor TWave.Create;
begin
  inherited Create;
  FSync := sndSync;
end;

destructor TWave.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TWave.Assign(Source: TPersistent);
begin
  if Source is TWave then
  begin
    Clear;

    FSize := TWave(Source).Size;
    if FSize > 0 then
    begin
      GetMem(FData, FSize);
      Move(TWave(Source).Data^, FData^, FSize);
      Changed;
    end;
  end
  else
    inherited Assign(Source);
end;

procedure TWave.LoadFromStream(Stream: TStream);
begin
  if Stream <> nil then
  begin
    Clear;

    FSize := Stream.Size;
    if FSize > 0 then
    begin
      GetMem(FData, FSize);
      Stream.Read(FData^, FSize);
    end;

    Changed;
  end;
end;

procedure TWave.SaveToStream(Stream: TStream);
begin
  if (Stream <> nil) and (FSize > 0) then
    Stream.Write(FData^, FSize);
end;

procedure TWave.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TWave.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, (FSize > 0));
end;

procedure TWave.LoadFromFile(const FileName: string);
var  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TWave.LoadFromResourceID(Instance: THandle; ResID: Integer;
  ResType: PChar);
var  Stream: TResourceStream;
begin
  Stream := TResourceStream.CreateFromID(Instance, ResID, ResType);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TWave.LoadFromResourceName(Instance: THandle;
  const ResName: string; ResType: PChar);
var  Stream: TResourceStream;
begin
  Stream := TResourceStream.Create(Instance, ResName, ResType);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TWave.SaveToFile(const FileName: string);
var  Stream: TFileStream;
begin
  Stream := TFileStream.Create(Filename, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TWave.ReadData(Stream: TStream);
begin
  LoadFromStream(Stream);
end;

procedure TWave.WriteData(Stream: TStream);
begin
  SaveToStream(Stream);
end;

function TWave.Flags: DWORD;
begin
  if FSync = sndSync then
    Result := SND_SYNC
  else
    Result := SND_ASYNC;
  if (sndLoop in FOptions) then Result := Result or SND_LOOP;
  if (sndNoDefault in FOptions) then Result := Result or SND_NODEFAULT;
  if (sndNoStop in FOptions) then Result := Result or SND_NOSTOP;
  if (sndNoWait in FOptions) then Result := Result or SND_NOWAIT;
end;

class function TWave.IsSoundInstalled: Boolean;
begin
  Result := waveOutGetNumDevs > 0;
end;

function TWave.Play: Boolean;
begin
  if (FSize = 0) or (not IsSoundInstalled) then
    Result := False
  else
    Result := PlaySound(FData, 0, Flags or SND_MEMORY);
end;

function TWave.PlayFile(const FileName: string): Boolean;
begin
  Result := PlaySound(PChar(FileName), 0, Flags or SND_FILENAME);
end;

function PlayResource(lpName, lpType: PChar; hInst: HMODULE; fdwSound: DWORD): BOOL; stdcall;
var
  hResInfo: HRSRC;
  hRes: HGLOBAL;
  lpRes: Pointer;
begin
  if StrIComp(lpType, RT_WAVE) = 0 then
    Result := PlaySound(lpName, hInst, fdwSound or SND_RESOURCE)
  else
  begin
    hResInfo := FindResource(hInst, lpName, lpType);
    if hResInfo <> 0 then
    begin
      hRes := LoadResource(hInst, hResInfo);
      if hRes <> 0 then
      begin
        lpRes := LockResource(hRes);
        if lpRes <> nil then
        begin
          Result := sndPlaySound(lpRes, fdwSound or SND_MEMORY);
          UnlockResource(hRes);
        end
        else
          Result := False;
        FreeResource(hRes);
      end
      else
        Result := False;
    end
    else
      Result := False;
  end;
end;

function TWave.PlayResourseName(Instance: THandle; const ResName: string;
  ResType: PChar): Boolean;
begin
  Result := PlayResource(PChar(ResName), ResType, Instance, Flags);
end;

function TWave.PlayResourseID(Instance: THandle; ResID: Integer;
  ResType: PChar): Boolean;
begin
  Result := PlayResource(Pointer(ResID), ResType, Instance, Flags);
end;

function TWave.PlayAlias(const AliasName: string): Boolean;
begin
  Result := PlaySound(PChar(AliasName), 0, Flags or SND_ALIAS);
end;

procedure TWave.Stop;
begin
  if not IsSoundInstalled then
    PlaySound(nil, 0, 0);
end;

procedure TWave.Clear;
begin
  if FSize > 0 then
  begin
    FreeMem(FData, FSize);
    FData := nil;
    FSize := 0;
    Changed;
  end;
end;

{ TWavePlayer }

constructor TWavePlayer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FWave := TWave.Create;
  FWave.OnChange := WaveChange;
  FSync := sndSync;
end;

destructor TWavePlayer.Destroy;
begin
  FWave.Stop;
  FWave.Free;
  inherited Destroy;
end;

procedure TWavePlayer.Loaded;
begin
  inherited Loaded;
end;

function TWavePlayer.Play: Boolean;
begin
  FWave.Sync := FSync;
  FWave.Options := FOptions;
  if Assigned(FBeforePlay) then FBeforePlay(Self);
  Result := FWave.Play;
  if Assigned(FAfterPlay) then FAfterPlay(Self);
end;

function TWavePlayer.PlayFile(const FileName: string): Boolean;
begin
  FWave.Sync := FSync;
  FWave.Options := FOptions;
  if Assigned(FBeforePlay) then FBeforePlay(Self);
  Result := FWave.PlayFile(FileName);
  if Assigned(FAfterPlay) then FAfterPlay(Self);
end;

function TWavePlayer.PlayResourseName(Instance: THandle; const ResName: string;
  ResType: PChar): Boolean;
begin
  FWave.Sync := FSync;
  FWave.Options := FOptions;
  if Assigned(FBeforePlay) then FBeforePlay(Self);
  Result := FWave.PlayResourseName(Instance, ResName, ResType);
  if Assigned(FAfterPlay) then FAfterPlay(Self);
end;

function TWavePlayer.PlayResourseID(Instance: THandle; ResID: Integer;
  ResType: PChar): Boolean;
begin
  FWave.Sync := FSync;
  FWave.Options := FOptions;
  if Assigned(FBeforePlay) then FBeforePlay(Self);
  Result := FWave.PlayResourseID(Instance, ResID, ResType);
  if Assigned(FAfterPlay) then FAfterPlay(Self);
end;

function TWavePlayer.PlayAlias(const AliasName: string): Boolean;
begin
  FWave.Sync := FSync;
  FWave.Options := FOptions;
  if Assigned(FBeforePlay) then FBeforePlay(Self);
  Result := FWave.PlayAlias(AliasName);
  if Assigned(FAfterPlay) then FAfterPlay(Self);
end;

procedure TWavePlayer.Stop;
begin
  FWave.Stop;
end;

procedure TWavePlayer.SetWave(Value: TWave);
begin
  FWave.Assign(Value);
end;

procedure TWavePlayer.WaveChange(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

end.
